%==========================================================================
% Estimation of Floating BG/FG Model by Nonparametric Belief Propagation
% Main inference module
% 2011. 1. 29. Suha Kwak, POSTECH.
%==========================================================================





%% PRELIMINARY
addpath 'ILP'

% input video for test ====================================================
% test video selection
vpath = '..\Data\';
fprintf(1, ' Test videos                                  \n');
fprintf(1, ' 1. Figure skating                            \n');
fprintf(1, ' 2. American football                         \n');
fprintf(1, ' 3. Hopkins, car 1                            \n');
fprintf(1, ' 4. Hopkins, people 1                         \n');
fprintf(1, ' 5. Hopkins, people 2                         \n');
fprintf(1, ' 6. Cycle                                     \n');
vid = input(' Select one of the above videos : ');
if isempty(vid)
    vid = 1;
end
switch vid
    case 1      % for video 1
        vpath = [vpath, 'skating1.mat'];
    case 2      % for video 2
        vpath = [vpath, 'NFL.mat'];
    case 3      % for video 3
        vpath = [vpath, 'hopkins_car1.mat'];
    case 4      % for video 4
        vpath = [vpath, 'hopkins_people1.mat'];
    case 5      % for video 5
        vpath = [vpath, 'hopkins_people2.mat'];
    case 6      % for video 6
        vpath = [vpath, 'cycle1.mat'];
    otherwise   % for video 1 (with no choice)
        vid = 1;
        vpath = [vpath, 'skating1.mat'];
end
% the sequences of optical flow maps and naive motion segmentations
% - generated in advance (done by Lim.)
evalc(['load ', vpath]);


% image size
irow = size(Iseq, 1);
icol = size(Iseq, 2);

% path for saving results 
rpath = ['..\Result\', num2str(vid), '\'];
mkdir(rpath);


% parameters ==============================================================
% threshold - the minimum aggregation weights (temporal model propagation)
minAW = 5e-2;

% block specification -----------------------------------------------------
% the number of rounds for iterative model-label estimation
nround = 2;

% how many pixels are needed to be an observation of the block
minpratio = 1e-1;

% block size
brow = 12;
bcol = 16;
        

% the parameters for motion estimation ------------------------------------
    % minpratio: how many pixels are needed to be an observation of a block
    % knum     : the number of kernel points for KDE (observation)
    % kstd     : (minimum) bandwidth of the kernels for KDE (observation)
    % cstd     : bandwidth of the compatibility potentials for NBP
    %            (cstd = 0 <=> do not use NBP)
    % maxrnd   : the maximum number of rounds for NBP
    % epsilon  : convergence threshold for NBP
    % autobw   : whether bandwidths are set manually or not 
    %            - true  = use the bandwidths automatically selected by MSP
    %            - false = use the bandwidths set manually

Pv_BG.minpratio = minpratio;
Pv_BG.knum = 50;
Pv_BG.kstd = 1;
Pv_BG.cstd = 0.5;
Pv_BG.maxrnd = 10;
Pv_BG.epsilon = 1e-10;
Pv_FG.minpratio = minpratio;
Pv_FG.knum = 50;
Pv_FG.kstd = 1;
Pv_FG.cstd = 3;
Pv_FG.maxrnd = 2;
Pv_FG.epsilon = 1e-10;


% the parameters for temporal propagation (prediction) --------------------
    % browstd  : bandwidth of normal kernel approximating block region 
    % bcolstd  : bandwidth of normal kernel approximating block region 
    % transtd  : BW of the state transition probability (Gaussian conv.)
    % minAW    : the minimum aggregation weights (temporal model propagation)

Pt.browstd = 4.5;        
Pt.bcolstd = 6.0;
Pt.transtd = 3;
Pt.minAW = minAW;


% the parameters for appearance estimation --------------------------------
    % epsilon  : for GMM product sampling
    % minpratio: how many pixels are needed to be an observation of a block
    % knum     : the (maximum) number of kernel points for KDE (observation)
    % kstd     : (minimum) bandwidth of the kernels for KDE (observation)
    % temp     : temperature parameter to determine bandwidth
    % nghb     : (obs) comparative weight to consider the neighboing blocks
    
Pc_BG.epsilon = 1e-10;
Pc_BG.minpratio = minpratio;
Pc_BG.knum = 150;
Pc_BG.kstd = 10;
Pc_BG.temp = 0.1;
Pc_BG.nghb = 0.5;
Pc_FG.epsilon = 1e-10;
Pc_FG.minpratio = minpratio;
Pc_FG.knum = 150;
Pc_FG.kstd = 10;
Pc_FG.temp = 0.05;
Pc_FG.nghb = 0.3;


% the parameters for pixel-wise label estimation --------------------------
    % nrnd     : the number of looopy BP rounds 
    % cvar     : bandwidth (variance) for compatibility clique potential
    % tempa    : temperature parameter for appearance observation likelihoods
    % tempm    : temperature parameter for motion observation likelihoods
    % epsc     : the lower bound of compatibility clique potentials
    % epso     : the lower bound of observation clique potentials
    % conv     : std for motion model convolution (motion likelihood)
    
Pl.nrnd = 50;
Pl.cvar = 25;
Pl.tempa = 0.9;
Pl.tempm = 0.1;
Pl.epsc = 0.15;
Pl.epso = 0.15;
Pl.conv = 3;


% preparation =============================================================
% specification of the image grid (B) -------------------------------------
    % brow      : block height
    % bcol      : block width
    % bnum_row  : the number of blocks at row-direction
    % bnum_col  : the number of blocks at column-direction
    % bnum      : the number of the entire blocks (bnum_row x bnum_col)
    % bcoord    : the coordinates of the blocks = [minR, minC, maxR, maxC]
    % G         : the pairwise graph structure (a binary matrix)

% block specification
B.brow = brow;
B.bcol = bcol;

% the number of blocks
B.bnum_row = ceil(irow / B.brow);
B.bnum_col = ceil(icol / B.bcol);
B.bnum = B.bnum_row * B.bnum_col;

% coordinates of each grid block
B.bcoord = zeros(B.bnum, 4);
for ridx = 1 : B.bnum_row
    for cidx = 1 : B.bnum_col        
        % coordinate of a block: minR, minC, maxR, maxC
        bidx = sub2ind([B.bnum_row, B.bnum_col], ridx, cidx);   % block index
        B.bcoord(bidx, 1 : 2) = [B.brow * (ridx - 1) + 1, ...
                                 B.bcol * (cidx - 1) + 1];
        B.bcoord(bidx, 3 : 4) = B.bcoord(bidx, 1 : 2) + [B.brow - 1, B.bcol - 1];
        if B.bcoord(bidx, 3) > irow
            B.bcoord(bidx, 3) = irow;
        end
        if B.bcoord(bidx, 4) > icol
            B.bcoord(bidx, 4) = icol;
        end
    end
end

% graph structure
B.G = false(B.bnum);
for ridx = 1 : B.bnum_row
    for cidx = 1 : B.bnum_col
        bidx = sub2ind([B.bnum_row, B.bnum_col], ridx, cidx);          % block index
        % link with the neighboring lower block
        if ridx < B.bnum_row
            bnidx = sub2ind([B.bnum_row, B.bnum_col], ridx + 1, cidx); % neighboring block index
            B.G(bidx, bnidx) = true;
            B.G(bnidx, bidx) = true;
        end
        % link with the neighboring right block
        if cidx < B.bnum_col
            bnidx = sub2ind([B.bnum_row, B.bnum_col], ridx, cidx + 1); % neighboring block index
            B.G(bidx, bnidx) = true;
            B.G(bnidx, bidx) = true;
        end
    end
end

% save parameters and the associated information
save([rpath, 'param.mat'], 'nround', 'minAW', 'B', 'Pv_BG', 'Pv_FG', 'Pt', ...
                                                   'Pc_BG', 'Pc_FG', 'Pl');




%% INITIAL STEP
tidx = 1;
disp('-- Frame 1')

% naive motion segmentation at the first frame
%   - already done by Lim.
I = double(Iseq(:, :, :, tidx));
L = BBP(Lseq(:, :, tidx));

% block appearance models at the first frame
disp('Initial appearance model estimation')
Qc_BG = AppModelEst(I, ~L, cell(1, B.bnum), B, Pc_BG, false);
Qc_FG = AppModelEst(I, L, cell(1, B.bnum), B, Pc_FG, true);

% saving results at each frame --------------------------------------------
% binary label map
imwrite(L, [rpath, 'L', num2str(tidx, '%03d'), '.jpg'], 'jpeg');

% interim results 
save([rpath, num2str(tidx), '.mat'], 'I', 'L', 'Qc_BG', 'Qc_FG');





%% ITERATIVE ESTIMATION
if(matlabpool('size') == 0)
    matlabpool('open');
end
for tidx = 2 : nframe
% initialization ----------------------------------------------------------
    % observations
    I = double(Iseq(:, :, :, tidx));    % input image
    O = Oseq(:, :, :, tidx);        	% optical flow map
    
    % initial label estimation
    L = BBP(Lseq(:, :, tidx));
    L(1 : B.brow, :) = false;
    L(irow - B.brow + 1 : irow, :) = false;
    L(:, 1 : B.bcol) = false;
    L(:, icol - B.bcol + 1 : icol) = false;

% iterative model-label estimation ----------------------------------------
    for ridx = 1 : nround
        disp(['-- Frame ', num2str(tidx), ' (round ', num2str(ridx), ')'])
        
        % motion model estimation
        disp('Motion model estimation')
        Qv_BG = MotModelEst(O, ~L, B, Pv_BG);
        Qv_FG = MotModelEst(O, L, B, Pv_FG);
        
        % temporal model propagation
        disp('Temporal model propagation')
        Qc_BG_tp = TempModelPropagate(Qv_BG, Qc_BG, B, Pt);
        Qc_FG_tp = TempModelPropagate(Qv_FG, Qc_FG, B, Pt);
        
        % label estimation
        disp('Label estimation')
        L = LabelEst(I, Qc_BG_tp, Qc_FG_tp, O, Qv_BG, Qv_FG, B, Pl);
        
        % consistent region (boundary is always BG)
        L(1 : B.brow, :) = false;
        L(irow - B.brow + 1 : irow, :) = false;
        L(:, 1 : B.bcol) = false;
        L(:, icol - B.bcol + 1 : icol) = false;
        
% saving results at each frame and each round -----------------------------
        % binary label map
        imwrite(L, [rpath, 'L', num2str(tidx, '%03d'), 'r', num2str(ridx), '.jpg'], 'jpeg');

%         % interim results
%         save([rpath, num2str(tidx), '_r', num2str(ridx), '.mat'], ...
%                     'I', 'O', 'L', 'Qv_BG', 'Qv_FG', 'Qc_BG_tp', 'Qc_FG_tp');
    end
    
    % appearance model estimation & update
    disp(['-- Frame ', num2str(tidx), ' (finished)'])
    disp('Appearance model update')
    Qc_BG = AppModelEst(I, ~L, Qc_BG_tp, B , Pc_BG, false);
    Qc_FG = AppModelEst(I, L, Qc_FG_tp, B, Pc_FG, true);
    
%     % final result, filtered appearance model only
%     save([rpath, num2str(tidx), '_f.mat'], 'Qc_BG', 'Qc_FG');
end




%% FINISH
matlabpool close;
clear;
rmpath 'ILP'


